import {Metadata, Http, DateHelper, Loader} from '@zeedhi/core';
import {
    Modal,
    ModalService,
    Form,
    DialogService,
    LoadingService,
    Select,
    SelectMultiple,
    DateRange,
    Col,
    Container,
    TreeGrid
} from '@zeedhi/common';
import { TekGrid } from "@zeedhi/teknisa-components-common";
import { CustomAlertService } from '@/Services/CustomAlertService';
import { Utilities } from '@/Services/Utilities';
import { ExpressionInterpreter } from '@/Services/ExpressionInterpreter';
import { PeriodsService } from '@/Services/PeriodsService';
import { Highlight } from '@zeedhi/zd-highlight-common';
import Prism from 'prismjs';
import 'prismjs/components/prism-markup-templating.min';
import 'prismjs/components/prism-php.min';
import {DSImportHistoryController} from "@/controllers/DataSource";

export class DataSourceController {
    private importHistoryModal!: Modal;
    public importHistoryData: any = {};

    public async openImportHistoryModalModal({ row }: any) {
        this.importHistoryModal?.destroy();
        this.importHistoryModal = await ModalService.createFromJson('dataSource/importHistory/importHistoryLog', true);
        this.importHistoryData = row || {};
        this.importHistoryModal.show();
    }
    public closeImportHistoryModal() {
        this.importHistoryModal.hide();
    }
    public importHistoryLogOnMounted({ component }: any) {
        const importHistoryLog = component.getChildInstance("importHistoryLog");
        importHistoryLog.value = this.importHistoryData.logs?.map((el: any) => `Type: ${el.type}\nMessage: ${el.message}`).join("\n\n");
    }
    public deleteImportHistory(params: any) {
        Utilities.genericBeforeDeleteTekGrid(params, "/importHistory/remove", "importHistory-grid");
    }
    public updateImportHistoryFilter() {
        if (Metadata.getInstances('importHistory-grid').length === 0 || !Metadata.getInstances('importHistory-grid')[0]) {
            return;
        }
        const dataSourceId = this.datasourceData._id;
        const importHistoryGrid: any = Metadata.getInstance<TekGrid>("importHistory-grid");
        importHistoryGrid.datasource.dynamicFilter = {
            dataSourceId: [
              {
                operation: "EQUALS",
                relation: "AND",
                value: dataSourceId,
              }
            ]
          };
        importHistoryGrid.datasource.get();
    }

    public importNowModal!: Modal;
    public importNowFormMounted({ component }: any) {
        const period = component.getChildInstance('periodNow');
        // Diminui o tamanho do menu de data
        const periodDropdown = period.childrenInstances[0].childrenInstances[0];
        periodDropdown.maxHeight = 190;
        periodDropdown.value = this.datasourceData.periodsCode;
        this.generatePeriodHelperOptions(period);
    }
    public closeImportNow() {
        this.importNowModal.hide();
        this.importNowModal.destroy();
    }
    public async importSchedulingNow() {
        this.importNowModal?.destroy();
        this.importNowModal = await ModalService.createFromJson('dataSource/ImportScheduling/importNowModal', true);
        this.importNowModal.show();
    }
    public async handleImportNow() {
        const importNowForm = Metadata.getInstance<Form>('importNowForm');
        const period = importNowForm.getChildInstance<DateRange>('periodNow');
        const data = this.generateImportScheduling({
            persistType: importNowForm.value.persistType,
            period: period.helperValue || importNowForm.value.period,
        });
        data.dataSource = this.datasourceData._id;
        data.status = 'queue';
        data.type = 'manualBI';
        if (!importNowForm.validate()) {
            return;
        }
        LoadingService.show();
        const { inProcess, processes } = await this.checkImportExistence();
        LoadingService.hide();
        if (inProcess) {
            DialogService.show({
                name: 'inProcess_dialog',
                type: 'warning',
                title: 'Atenção',
                text: 'Há um ou mais imports em progresso. O que você gostaria de fazer?',
                buttons: [
                    { name: 'replace', component: 'ZdButton', label: 'Substituir processos de importação', events: { click: () => { DialogService.hide(); this.replaceImportProcess(processes, data); }}, outline: true },
                    { name: 'add', component: 'ZdButton', label: 'Adicionar a fila', events: { click: () => { DialogService.hide(); this.addScheduleToQueue(data); }}, outline: true },
                    { name: 'cancel', component: 'ZdButton', label: 'Cancelar', events: { click: () => { DialogService.hide(); this.closeImportNow(); }} }
                ]
            });
        } else {
            await this.saveImportNowSchedule(data);
        }
    }
    public async checkImportExistence() {
        const res = await Http.post("/dataSource/findImports", {
			"requestType": "FilterData",
			"filter": [
				{
					"name": "dataSourceId",
					"operator": "=",
					"value": this.datasourceData._id
				}
			]
		});
        const processes = res.data.dataset;
        return {
            processes,
            inProcess: processes.totalImportsInProcess[0] > 0
        };
    }
    public async saveImportNowSchedule(importScheduling: any) {
        let status = false;
        let errorMessage = '';
        const res = await Http.post("/importSchedule/save", {
            "requestType": "Row",
            "row": { ...importScheduling }
        });
        if (res.data.notifications && res.data.notifications[0]) {
            const notification = res.data.notifications[0];
            if (notification.type === 'success') {
                status = true;
            } else {
                errorMessage = notification.message;
            }
        } else if (res.data.messages && res.data.messages[0]) {
            const message = res.data.messages[0];
            if (message.type === 'E') {
                errorMessage = message.message;
            } else {
                status = true;
            }
        }
        LoadingService.hide();
        if (status) {
            CustomAlertService.successAlert("Importation successfully scheduled! You can follow up the process at the tab 'Import History' in a few moments");
            this.closeImportNow();
        } else {
            CustomAlertService.errorAlert(errorMessage || "Um erro ocorreu ao realizar a importação");
        }
    }
    public async replaceImportProcess(processes: any, importScheduling: any) {
        if (processes.totalImportsBlocked[0] > 0) {
            CustomAlertService.errorAlert('IMPORT_IN_PROCESS');
        } else {
            LoadingService.show();
            await Http.post("/dataSource/cancelImportation", {
                "requestType": "FilterData",
                "filter": [
                    {
                        "name": "dataSourceId",
                        "operator": "=",
                        "value": this.datasourceData._id
                    }
                ]
            });
            await this.saveImportNowSchedule(importScheduling);
        }
    }
    public async addScheduleToQueue(scheduleItem: any) {
        LoadingService.show();
        const userRes = await Http.post("/getLoggedUser", {nrorg: null, nrorgPrincipal: null});
        const { username } = userRes.data;
        scheduleItem.userResponsible = username;
        await Http.post("/manualImportQueue/save", {
			"requestType": "Row",
			"row": {
                'scheduleItem': scheduleItem
            }
		});
        LoadingService.hide();
        this.closeImportNow();
    }

    public async generatePeriodHelperOptions(period: any) {
        const res = await Http.post("/periods");
        const data = res.data.data;
        if (period.helperOptions && period.helperOptions.length > 0) {
            period.helperOptions.forEach((option: any) => DateHelper.unregister(option.periodsCode));
        }
        data.forEach((option: any) => DateHelper.register(option.periodsCode, option.name, () => {
            const service = new PeriodsService();
            const periods = service.decodePeriod(option.periodsCode);
            return [periods.start, periods.end].map((date: any) => date.split('/').reverse().join('-')) as any;
        }));
        period.helperOptions = data.map((option: any) => option.periodsCode).filter((val: any) => val);
        period.helperValue = this.datasourceData.period;
        period.updateHelperHint();
    }
    public generateImportScheduling(importNowParams: any = null) {
        const form = Metadata.getInstance<Form>('importScheduling');
        const afterImportScript = Metadata.getInstance<Highlight>('afterImportScript');
        const beforeImportScript = Metadata.getInstance<Highlight>('beforeImportScript');
        const period = form.getChildInstance<DateRange>('period');
        const values: any = {
            ...form.value,
            afterImportScript: afterImportScript.code,
            beforeImportScript: beforeImportScript.code,
            period: period.helperValue || (form.value.period ? { start: form.value.period[0], end: form.value.period[1] } : null),
            dateFilterColumn: form.value.dateFilterColumnLabel?.name,
            dateFilterColumnLabel: form.value.dateFilterColumnLabel?.label
        };
        if (importNowParams) {
            values.persistType = importNowParams.persistType || values.persistType;
            values.period = importNowParams.period || values.period;
        }
        return values;
    }
    public async saveImportScheduling() {
        const form = Metadata.getInstance<Form>('importScheduling');
        if (!form.validate()) {
            return;
        }
        const values = this.generateImportScheduling();
        console.log(values);
        Object.assign(this.datasourceData, values);
        LoadingService.show();
        try {
            const res = await this.saveDataSource(this.datasourceData);
            CustomAlertService.saveSuccessAlert();
            if (res.data.messages?.length > 0) {
                const message = res.data.messages[0].message;
                DialogService.show({
                  name: 'warning-dialog',
                  type: 'error',
                  title: ' ',
                  text: message
                });
            }
            this.importSchedulingUpdateForm();
        } catch (err) {
            LoadingService.hide();
            CustomAlertService.errorAlert("Ocorreu um erro.");
            console.error(err);
        }
    }
    public importSchedulingUpdateForm() {
        // Se formulário não foi montado, não faz nada
        if (Metadata.getInstances('importScheduling').length === 0 || !Metadata.getInstances('importScheduling')[0]) {
            return;
        }
        const form = Metadata.getInstance<Form>('importScheduling');
        const keys = form.getChildInstance<Select>('keys');
        const dateFilter = form.getChildInstance<Select>('dateFilterColumnLabel');
        const period = form.getChildInstance<DateRange>('period');
        const afterImportScript = Metadata.getInstance<Highlight>('afterImportScript');
        const beforeImportScript = Metadata.getInstance<Highlight>('beforeImportScript');
        keys.datasource.updateData(this.datasourceData.columns || []);
        dateFilter.datasource.updateData(this.datasourceData.columns?.filter((col: any) => col.isDate) || []);

        // Diminui o tamanho do menu de data
        const periodDropdown = period.childrenInstances[0].childrenInstances[0];
        periodDropdown.maxHeight = 190;
        this.generatePeriodHelperOptions(period);

        const ignoreKeys: string[] = ["period"];
        Object.keys(form.value).filter((key: string) => !ignoreKeys.includes(key)).forEach((key: any) => { form.value[key] = this.datasourceData[key]; });
        period.change();
        if (this.datasourceData.period) {
            if (typeof this.datasourceData.period === 'string') {
                const service = new PeriodsService();
                const values = service.decodePeriod(this.datasourceData.period);
                period.setDateValue(values.start + ' - ' + values.end);
                if (period.helperOptions?.includes(this.datasourceData.period)) {
                    period.helperValue = this.datasourceData.period;
                }
            } else {
                period.setDateValue(this.datasourceData.period.start + ' - ' + this.datasourceData.period.end);
            }
        } else {
            period.setDateValue("");
        }
        afterImportScript.code = this.datasourceData.afterImportScript || "";
        beforeImportScript.code = this.datasourceData.beforeImportScript || "";
    }
    public loadLanguageAfterImport() {
        const { code, language } = Metadata.getInstance<Highlight>('afterImportScript');
        return Prism.highlight(code, Prism.languages[language], language);
    }
    public loadLanguageBeforeImport() {
        const { code, language } = Metadata.getInstance<Highlight>('beforeImportScript');
        return Prism.highlight(code, Prism.languages[language], language);
    }

    private structureModal!: Modal;
    public structureData: any = {};
    public async openStructuresModal({ row }: any) {
        this.structureModal?.destroy();
        this.structureModal = await ModalService.createFromJson('dataSource/RelateStructures/RelateStructuresModal', true);
        this.structureData = row || {};
        this.structureModal.show();
    }
    public closeStructuresModal() {
        this.structureModal.hide();
    }
    public structureModalMounted({ component }: any) {
        const listEl = component;
        const relationsValues = this.datasourceData.relatedStructuresType.find((el: any) => el.structureType === this.structureData._id)?.relations || [];
        const relations = this.structureData.fieldKey.map((key: string) => ({ key, value: relationsValues.filter((relation: any) => relation.structureColumn === key).map((el: any) => el.dataSourceColumn) }));
        listEl.children = relations.map(this.generateRelationListItem);
    }
    public generateRelationListItem = (relation: any, id: number) => ({
        "name": "relation_" + relation.key,
        "component": "ZdRow",
        "dense": false,
        "cssClass": "zd-justify-space-between zd-align-center zd-py-4",
        "children": [
            {
                "name": "relationName_" + id,
                "component": "ZdText",
                "text": relation.key,
                "readonly": true,
                "showLabel": false,
                "showHelper": false,
                "cssClass": "zd-col-4",
                "disabled": true
            },
            {
                "name": "relationSelect_" + id,
                "component": "ZdSelectMultiple",
                "value": relation.value,
                "showHelper": false,
                "showLabel": false,
                "disabledItems": "{{DataSourceController.structureRelationSelectDisabled}}",
                "cssClass": "zd-col-7",
                "dataText": ["label"],
                "dataValue": "name",
                "datasource": {
                    "type": "memory",
                    "uniqueKey": "_id",
                    "data": this.datasourceData.relations.map(({name, label}: any) => ({name, label})) || []
                },
                "events": {
                    "focus": "{{DataSourceController.structureRelationSelectFocus}}"
                }
            }
        ]
    });
    public currentStructureRelationSelect: string = "";
    public structureRelationSelectFocus({ component }: any) {
        this.currentStructureRelationSelect = component.parent.name.split(/_(.+)/)[1];
    }
    public get structureRelationSelectDisabled() {
        const list = Metadata.getInstance<Container>("structureDetailList");
        const values = list.childrenInstances?.map((el: any) => ({name: el.name.split(/_(.+)/)[1], value: el.childrenInstances[1]?.value})) || [];
        return values.filter((el: any) => el.name !== this.currentStructureRelationSelect).map((el: any) => el.value).flat();
    }
    public async saveStructuresModal() {
        const list = Metadata.getInstance<Container>("structureDetailList");
        const values = list.childrenInstances?.map((el: any) => ({name: el.name.split(/_(.+)/)[1], value: el.childrenInstances[1]?.value})) || [];
        let dataHolder = this.datasourceData.relatedStructuresType.find((el: any) => el.structureType === this.structureData._id);
        if (!dataHolder) {
            dataHolder = { structureType: this.structureData._id, relations: [] };
            this.datasourceData.relatedStructuresType.push(dataHolder);
        }
        dataHolder.relations = values.map((el: any) => el.value.map((value: any) => ({structureColumn: el.name, dataSourceColumn: value})).flat()).flat();
        try {
            const res = await this.saveDataSource(this.datasourceData);
            if (res.data.messages?.length > 0) {
                const message = res.data.messages[0].message;
                DialogService.show({
                  name: 'warning-dialog',
                  type: 'error',
                  title: ' ',
                  text: message
                });
            } else {
                CustomAlertService.saveSuccessAlert();
            }
            this.closeStructuresModal();
        } catch (err) {
            LoadingService.hide();
            CustomAlertService.errorAlert("Ocorreu um erro.");
            console.error(err);
        }
    }

    public indexesData: any;
    private indexModal!: Modal;
    public get isAddIndex() {
        return this.indexesData === undefined;
    }
    public get isDetailIndex() {
        return this.indexesData !== undefined;
    }
    public get indexModalTitle() {
        return (this.indexesData ? "" : "Adicionar") + " Índice";
    }
    public get indexesFormGrid() {
        return this.isAddIndex ? {name: 9, column: 6} : {name: 12, column: 9};
    }
    public updateIndexesFilter() {
        // Se grid não foi montado, não faz nada
        if (Metadata.getInstances("indexes-grid").length === 0 || !Metadata.getInstances("indexes-grid")[0]) {
            return;
        }
        const dataSourceId = this.datasourceData._id;
        const indexesGrid = Metadata.getInstance<TekGrid>("indexes-grid");
        indexesGrid.datasource.setFilter({ dataSourceId });
    }
    public async openIndexesModal({ row }: any) {
        this.indexModal?.destroy();
        this.indexModal = await ModalService.createFromJson('dataSource/Indexes/IndexDetailModal', true);
        this.indexesData = row || undefined;
        this.indexModal.show();
    }
    public closeIndexesModal() {
        this.indexModal.hide();
        this.indexModal.destroy();
    }
    public indexesColumnListHolderMounted() {
        if (this.indexesData) {
            const columnList = Metadata.getInstance<Col>("indexesColumnList");
            const indexesForm = Metadata.getInstance<Form>("indexesForm");
            const {name, key} = this.indexesData;
            indexesForm.value.indexName = name;
            Object.keys(key).forEach((column: string) => {
                this.newColumnOnIndexForm({column, order: key[column] === 1 ? "Crescente" : "Decrescente"});
            });
            // Remove primeiro elemento que vem como padrão
            columnList.childrenInstances.splice(0, 1);
            columnList.children.splice(0, 1);
        }
    }
    public async columnListOnIndexFormMounted({ component }: any) {
        const row = component.childrenInstances[0];
        const val = component.children[0].value;
        await row.datasource.updateData(this.datasourceData.columns);
        if (val) {
            row.setValue(val);
        }
    }
    public newColumnOnIndexForm(data: any) {
        const columnList = Metadata.getInstance<Col>("indexesColumnList");
        const usedNumbers = columnList.children.map((params: any) => Number(params.name.split('_')[1])).sort((a: number, b: number) => (a > b ? 1 : -1));
        let newNumber = usedNumbers.length;
        for (let i = 0; i < usedNumbers.length; i += 1) {
            if (usedNumbers[i] !== i) {
                newNumber = i;
                break;
            }
        }
        const template = JSON.parse(JSON.stringify({...columnList.children[0]}));
        template.name = "columnRow_" + newNumber;
        template.children = template.children?.map((obj: any) => ({...obj, value: "", name: obj.name.split('_')[0] + '_' + newNumber })) || [];
        if (data && data.column && data.order) {
            template.children[0].value = data.column;
            template.children[1].value = data.order;
        } else {
            template.children[1].value = "Crescente";
        }
        columnList.children.push(template);
        return template;
    }
    public removeColumnOnIndexForm({ component }: any) {
        const columnList = Metadata.getInstance<Col>("indexesColumnList");
        const column = component.parent;
        const index = columnList.childrenInstances.findIndex((el: any) => el.name === column.name);
        if (index !== -1 && columnList.children.length > 1) {
            columnList.childrenInstances.splice(index, 1);
            columnList.children.splice(index, 1);
        }
    }
    public confirmSaveIndexes() {
        const indexesForm = Metadata.getInstance<Form>("indexesForm");
        const formsHolder = Metadata.getInstance<Col>("indexesColumnList");
        //  A utilização do map permite que o metódo validate seja chamado em todos os formulários, diferente do every que pararia no primeiro false
        if (![indexesForm, ...formsHolder.childrenInstances].map((form: any) => form.validate()).every((valid: boolean) => valid)) {
            return;
        }
        DialogService.show({
            name: "indexSaveConfirm",
            type: "warning",
            title: "Confirmação",
            text: "Em fontes de dados muito extensas, o processo de criação de índices pode demorar muito tempo. A fonte de dados ficará inacessível até o término deste processo e não será possível interrompê-lo.<br><br>Deseja continuar?",
            buttons: [
                { name: 'não', component: 'ZdButton', label: 'Não', outline: true, events: { click: () => DialogService.hide() } },
                { name: 'sim', component: 'ZdButton', label: 'Sim', events: { click: () => { this.saveIndexesForm(); DialogService.hide(); } } }
            ],
        });
    }
    public async saveIndexesForm() {
        const indexesForm = Metadata.getInstance<Form>("indexesForm");
        const name = indexesForm.value.indexName;
        const formsHolder = Metadata.getInstance<Col>("indexesColumnList");
        const columnsValue = formsHolder.childrenInstances.map((params: any) => {
            const number = Number(params.name.split('_')[1]);
            return { direction: params.value["order_" + number] === "Crescente" ? 1 : -1, name: params.value["column_" + number]};
        });
        const data = {
            name,
            dataSource: this.datasourceData._id,
            COLUMN_: columnsValue.map((column: any) => column.name),
            ORDER_: columnsValue.map((column: any) => column.direction)
        };
        try {
            LoadingService.show();
            const res = await Http.post("/dataSource/saveIndexes", {
                requestType: "Row",
                row: data
            });
            LoadingService.hide();
            if (res.error) {
                CustomAlertService.errorAlert(res.error);
            } else {
                CustomAlertService.saveSuccessAlert();
                this.closeIndexesModal();
                this.reloadIndexesGrid();
            }
        } catch (e) {
            LoadingService.hide();
            CustomAlertService.errorAlert(String(e));
            console.error(e);
        }
    }
    public async deleteIndexes(params: any) {
        Utilities.genericBeforeDeleteTekGrid(params, "/dataSource/deleteIndexes", "indexes-grid", "Row");
    }
    public reloadIndexesGrid() {
        const grid = Metadata.getInstance<TekGrid>("indexes-grid");
        grid.reload();
    }
    // Controle do Master Detail
    public isGridVisible: any = true;
    public get isDetailVisible() {
        return !this.isGridVisible;
    }
    public showGrid() {
        this.isGridVisible = true;
    }
    public showDetail() {
        this.isGridVisible = false;
    }
    //  Referente ao componente Highlight
    public get addDatasourceQuery() {
        const queries = Metadata.getInstances('query');
        return queries.find((query: any) => query.parent.name === 'modalDataSourceForm');
    }
    public get editDatasourceQuery() {
        const queries = Metadata.getInstances('query');
        return queries.find((query: any) => query.parent.name === 'dataSourceForm');
    }
    public loadLanguage = this.loadLanguageAux.bind(this);
    public loadLanguageAux() {
        const query = this.isGridVisible ? this.addDatasourceQuery : this.editDatasourceQuery;
        if (query) {
            const { code, language } = query;
            return Prism.highlight(code, Prism.languages[language], language);
        }
        return "";
    }
    public loadLanguageExpression(params: any) {
        const { code, language } = Metadata.getInstance<Highlight>('expression');
        return Prism.highlight(code, Prism.languages[language], language);
    }
    // Métodos de Dialog e Modal
    public showDeleteDialog(params: any) {
        Utilities.genericBeforeDeleteTekGrid(params, "/dataSource/remove", "datasource-grid");
    }

    // Aba Colunas
    private columnModal!: Modal;
    public columnModalData: any;
    public columnModalType: string = "";
    public get isColumnFact() {
        const form = Metadata.getInstance<Form>("columnForm");
        return form.value.type === "fact";
    }
    public get isColumnDimension() {
        const form = Metadata.getInstance<Form>("columnForm");
        return form.value.type === "dimension";
    }
    public get isColumnFactAndNotGenerated() {
        return this.isColumnFact && !this.columnModalData?.isGenerated;
    }
    public get isColumnUsingCase() {
        const form = Metadata.getInstance<Form>("columnForm");
        return this.isColumnFact && form.value.specialFunction === "case";
    }
    public get isColumnNotUsingCase() {
        const form = Metadata.getInstance<Form>("columnForm");
        return !this.isColumnFact || form.value.specialFunction !== "case";
    }
    public get isColumnUsingModule() {
        const form = Metadata.getInstance<Form>("columnForm");
        return this.isColumnFact && form.value.specialFunction === "mod";
    }
    public get isColumnUsingPower() {
        const form = Metadata.getInstance<Form>("columnForm");
        return this.isColumnFact && form.value.specialFunction === "power";
    }
    public async openColumnModal({ row }: any) {
        this.columnModal?.destroy();
        this.columnModal = await ModalService.createFromJson('dataSource/Columns/columnModal', true);
        this.columnModalData = row === undefined ? {type: 'fact', datatype: 'A', direction: 'asc', relatedFactMultiple: []} : row;
        this.columnModal.show();
    }
    public closeColumnModal() {
        this.columnModal.hide();
        this.columnModal.destroy();
    }
    public columnFormMounted({ component }: any) {
        const expression = Metadata.getInstance<Highlight>('expression');
        const relatedFacts = ['relatedFactMultiple', 'relatedFact', 'secondRelatedFact'];
        relatedFacts.forEach((relatedFactName: any) => {
            const relatedFact = Metadata.getInstance<SelectMultiple>(relatedFactName);
            relatedFact.datasource.updateData(this.datasourceData.columns.filter((col: any) => col.isFact));
            relatedFact.value = this.columnModalData[relatedFactName] || [];
        });
        component.value = this.columnModalData;
        expression.code = component.value.expression || "";
    }
    public columnSpecialFunctions: any = {};
    public columnFormGrid = { datatype: 3, color: 6};
    public updateColumnConditionalFormats() {
        const direction = Metadata.getInstance<Select>('direction');
        const specialFunction = Metadata.getInstance<Select>('specialFunction');
        const relatedFact = Metadata.getInstance<Select>('relatedFact');
        const relatedFact2 = Metadata.getInstance<Select>('secondRelatedFact');
        const potentialExpoent = Metadata.getInstance<Select>('potentialExpoent');

        if (this.isColumnFact) {
            direction.addValidation('required', {});
            this.columnFormGrid.datatype = 6;
            this.columnFormGrid.color = 3;
        } else {
            direction.removeValidation('required');
            this.columnFormGrid.datatype = 3;
            this.columnFormGrid.color = 6;
        }

        const requiresRelatedFact = ['abs', 'sign', 'ceil', 'floor', 'sqrt', 'mod', 'power'];
        if (specialFunction.value === 'power') {
            potentialExpoent.addValidation('required', {"message": "Por favor, preencha o expoente"});
        } else {
            potentialExpoent.removeValidation('required');
        }

        if (specialFunction.value === 'module') {
            relatedFact2.addValidation('required', {"message": "Por favor, preencha o segundo fato relacionado para a função mod"});
        } else {
            relatedFact2.removeValidation('required');
        }

        if (requiresRelatedFact.includes(specialFunction.value)) {
            relatedFact.addValidation('required', {"message": 'Por favor, preencha o fato relacionado para a função ' + specialFunction.value});
        } else {
            relatedFact.removeValidation('required');
        }
    }
    public async importColumns() {
        const gridColumns = Metadata.getInstance<TekGrid>("gridColumns");
        try {
            gridColumns.datasource.loading = true;
            const res = await Http.post("/dataSource/syncColumns", {
                requestType: "Row",
                row: this.datasourceData || {}
            });
            gridColumns.datasource.loading = false;
            if (res.data.messages?.length > 0) {
                const message = res.data.messages[0].message;
                DialogService.show({
                  name: 'warning-dialog',
                  type: 'warning',
                  title: ' ',
                  text: message
                });
                if (res.data.messages[0].type === 'E') {
                    return;
                }
            }
            const columns = res.data.dataset && res.data.dataset.columns ? res.data.dataset.columns : [];
            this.datasourceData.columns = columns;
            this.columnGridChange({component: gridColumns});
            gridColumns.datasource.loading = false;
        } catch (err) {
            gridColumns.datasource.loading = false;
            CustomAlertService.errorAlert("Ocorreu um erro.");
            console.error(err);
        }
    }

    public selectColumn({ row, component }: any) {
        if (row?.isGenerated) {
            DialogService.show({
              name: 'warning-column-select',
              type: 'warning',
              title: ' ',
              text: 'Não é possivel selecionar uma coluna gerada!'
            });
            component.selectedRows = component.selectedRows.filter((value: any) => !value.isGenerated);
        }
    }
    public async deleteColumn({ component }: any) {
        const columnsToDelete = component.selectedRows.filter((value: any) => !value.isGenerated);
        if (columnsToDelete.length === 0) {
            return;
        }
        this.datasourceData.columns = this.datasourceData.columns.filter((column: any) => !columnsToDelete.includes(column));
        const res = await this.saveColumns(this.datasourceData);
        const columnGrid = Metadata.getInstance<TekGrid>("gridColumns");
        columnGrid.datasource.updateData(res.columns);
        CustomAlertService.deleteSuccessAlert();
    }

    public testExpression() {
        let expression: any = "";
        const formValues = Metadata.getInstance<Form>("columnForm").value;
        const expressionEl = Metadata.getInstance<Highlight>("expression");
        try {
            if (formValues.specialFunction === "case") {
                expression = this.readCaseExpression(expressionEl.code, this.datasourceData.hasDataMining, formValues);
            } else {
                expression = this.readRegularExpression(formValues.name, expressionEl.code);
            }
            if (expression || expressionEl.code) {
                if (expression) {
                    CustomAlertService.successAlert("Expressão válida.");
                } else {
                    CustomAlertService.errorAlert("Expressão inválida.");
                }
            }
            return expression;
        } catch (e) {
            console.error(e);
            CustomAlertService.errorAlert(String(e));
        }
        return undefined;
    }
    public readRegularExpression(name: string, code: string) {
        const data = this.datasourceData.columns;
        const facts = data.filter((column: any) => column.isFact === true).map((column: any) => column.name);
		const nonFacts = data.filter((column: any) => column.isFact === false).map((column: any) => column.name);
		return ExpressionInterpreter.readExpression(code, facts, nonFacts, name);
    }
    public readCaseExpression(expression: string, hasDataMining: any, formValues: any) {
		if (expression) {
            if (!hasDataMining) {
                if (this.checkExpressionString(expression.toUpperCase())) {
					return expression;
                }
                CustomAlertService.errorAlert("Para criar uma expressão case sem mineração de dados, você deve consultar a sintaxe \"CASE\" de seu banco de dados.");
            } else {
                const caseExp = this.formatCaseExp(expression);
                const regEx = new RegExp("(CASE.* [^ ].* [<>!=].* [^ ].* (THEN).* [^ ].*).* DEFAULT.* ([^ ].*)");
                const testExp = regEx.test(caseExp);
                if (testExp) {
                    formValues.caseExp = caseExp;
                    return expression;
                }
                CustomAlertService.errorAlert("Para expressões CASE, siga o modelo: CASE [condição] THEN [expressão] DEFAULT [expressão]\n\nEx: CASE $fatoRelacionado1 != 0 THEN (( $fatoRelacionado2 - $fatoRelacionado1 )*100 / $fatoRelacionado1 ) DEFAULT 100");
            }
		} else {
            CustomAlertService.errorAlert("O campo \"expressões\" é obrigatório para funções CASE.");
		}
		return "";
    }
    public checkExpressionString(expression: string) {
		const arrayKeyWords = expression.split(" ");
		const arrayExpression = expression.split("");
		let checkExpression;
		if (arrayKeyWords[0] === "CASE") {
            let countChar = 0;
            arrayExpression.forEach((val: string) => {
                if (val === '(') {
					countChar += 1;
				} else if (val === ')') {
					countChar -= 1;
					if (countChar < 0) {
						CustomAlertService.errorAlert("Não há \"(\" para fechar com \")\".");
						checkExpression = false;
					}
				}
            });
			if (countChar > 0) {
				CustomAlertService.errorAlert("Existem mais \"(\" do que o necessário.");
				checkExpression = false;
			} else if (countChar < 0) {
				CustomAlertService.errorAlert("Existem mais \")\" do que o necessário.");
				checkExpression = false;
			} else if (countChar === 0 && checkExpression !== false) {
				checkExpression = true;
				CustomAlertService.errorAlert("Por favor, verifique a sintaxe da expressão case do banco de dados usado.");
			}
		} else {
			checkExpression = false;
		}
		return checkExpression;
	}
	public formatCaseExp(expression: string) {
        let newExp = '';
		const words = expression.split(' ');

		words.forEach((word: string) => {
			if (word[0] !== '$') {
				word = word.toUpperCase();
			}

			word = word.replace('\n', ' ');
			newExp += word + ' ';
		});

		return newExp;
	}

    public async saveColumnModal() {
        const form = Metadata.getInstance<Form>("columnForm");
        const query = Metadata.getInstance<Highlight>('expression');
        const values = {...form.value};
        const queryCode = query.code ? this.parseQueryCode(query.code) : "";
        if (!form.validate() || (this.isColumnFactAndNotGenerated && this.isColumnUsingCase && !queryCode.trim())) {
            if (!queryCode.trim()) {
                CustomAlertService.errorAlert("O campo de código é obrigatório e deve ser preenchido");
            }
            return;
        }
        const aux = {
            ...this.columnModalData,
            ...values,
            expression: queryCode,
            isDate: values.datatype === 'D',
            isDimension: values.type === 'dimension',
            isFact: values.type === 'fact',
            isGenerated: this.columnModalData.isGenerated || false
        };
        if (aux.color && !Array.isArray(aux.color)) {
            aux.color = [aux.color];
        }
        let ignoreAttrs: string[] = [];
        if (aux.isFact) {
            ignoreAttrs = aux.isGenerated ? ["isKey", "isFunctionAccumulative", "specialFunction", "relatedFactLabel", "expression", "validateExp"] : ["isKey"];
        } else if (aux.isDimension) {
            ignoreAttrs = ["function", "isFunctionAccumulative", "specialFunction", "relatedFactLabel", "expression", "validateExp", "mask"];
        }
        ignoreAttrs.forEach((key: string) => { delete aux[key]; });
        // Caso a coluna já existisse, substitui a posição pelas novas edições na coluna, caso o contrario, a insere no vetor
        const oldPos = this.datasourceData.columns.findIndex((col: any) => col === this.columnModalData);
        if (oldPos >= 0) {
            this.datasourceData.columns[oldPos] = aux;
        } else {
            this.datasourceData.columns.push(aux);
        }
        this.columnModalData = aux;
        let res = null;
        if (this.columnModalData.isDimension) {
            res = await this.saveColumn(this.datasourceData, this.columnModalData);
        } else if (this.datasourceData.hasDataMining) {
            if (queryCode !== "") {
                if (this.testExpression() !== undefined || this.readRegularExpression(this.columnModalData.name, queryCode) !== undefined) {
					res = await this.saveColumn(this.datasourceData, this.columnModalData);
				} else {
					CustomAlertService.errorAlert("A coluna não pode ser salva.");
				}
            } else {
                res = await this.saveColumn(this.datasourceData, this.columnModalData);
            }
        } else if (this.readCaseExpression(this.columnModalData.expression, false, this.columnModalData) !== undefined || this.readRegularExpression(this.columnModalData.name, queryCode) !== undefined) {
			res = await this.saveColumn(this.datasourceData, this.columnModalData);
		} else {
			CustomAlertService.errorAlert("A coluna não pode ser salva.");
		}

        if (res) {
            const columnGrid = Metadata.getInstance<TekGrid>("gridColumns");
            columnGrid.datasource.updateData(res.columns);
            this.closeColumnModal();
        }
    }
    public saveColumn(datasourceData: any, columnData: any) {
		columnData._altered = true;
		if (!columnData.isGenerated && columnData.expression) {
			let expression = columnData.expression;
			const beginningExpression = expression.substring(0, 4).toUpperCase();
			if (beginningExpression !== "CASE") {
                expression = this.readRegularExpression(columnData.name, expression);
			}
			columnData.functions = expression.functions;
			columnData.mathExpressions = expression.mathExpressions;
		}
		return this.saveColumns(datasourceData);
    }
    public async saveColumns(datasourceData: any) {
        datasourceData._columnsChanged = true;
        datasourceData.columns = datasourceData.columns.map((column: any) => ({
            ...column,
            isSortable: !!column.direction,
            isFact: column.type === 'fact',
            isDimension: column.type === 'dimension'
		}));
		datasourceData.keys = datasourceData.columns.filter((column: any) => column.isKey).map((currentValue: any) => currentValue.name);
		const res = await this.saveDataSource(datasourceData);
		return {res, columns: datasourceData.columns};
    }

    // TRECHO DE CODIGO RELACIONADO AO MASTERDETAIL, GRID E FORM DO DATASOURCE
    private modal!: Modal;
    public closeDetails() {
        if (this.isDetailVisible) {
            this.showGrid();
        } else {
            this.modal.hide();
        }
    }
    public async createAddModal() {
        this.modal = await ModalService.createFromJson('dataSource/DataSource/AddModal', true);
    }

    public datasourceData: any;
    public detailTitleText: string = '';
    public switchToDetail({ component, row }: any) {
        if (component.editing === true) return;
        this.datasourceData = row;
        this.detailTitleText = this.datasourceData?.name || "Fonte de Dados";
        this.showDetail();
        this.fillDataSourceForm({ row });
        const columnGrid = Metadata.getInstances("gridColumns");
        if (columnGrid.length > 0 && columnGrid[0]) {
            this.columnGridChange({ component: columnGrid[0] });
        }
        this.updateIndexesFilter();
        this.importSchedulingUpdateForm();
        this.updateImportHistoryFilter();
    }
    public async openModal({ component, row }: any) {
        this.modal?.destroy();
        await this.createAddModal();
        this.datasourceData = {joinDataSource: "N", hasDataMining: "S", willUpdateIndicators: "N", isScheduleActive: "S"};
        this.modal.show();
    }
    public columnGridChange({ component }: any) {
        const columns = this.datasourceData.columns || [];
        component.datasource.updateData(columns);
    }
    // Métodos da query
    public extractStringsRangeFromQuery(code: string) {
        const stringRegex = /(["'])(?:(?=(\\?))\2.)*?\1/g;
        return [...code.matchAll(stringRegex)].map((codeString: any) => [codeString.index, codeString.index + codeString[0].length]);
    }

    public strReplaceNotIn(forbideenRange: any, inForbideenFunc: any, notInForbideenFunc: any) {
        return (...params: any) => {
            const commentStart = params[params.length - 2];
            const containingString = forbideenRange.find(([start, end]: any) => commentStart > start && commentStart < end);
            if (containingString) {
                return inForbideenFunc(params, containingString);
            }
            return notInForbideenFunc(params);
        };
    }

    public clearComments(code: string): string {
        if (code === "") {
            return "";
        }
        const commentRegex = /(--.*)|(((\/\*)+?[\w\W]+?(\*\/)+))/g;
        const stringRanges = this.extractStringsRangeFromQuery(code);
        return code.replace(commentRegex, this.strReplaceNotIn(stringRanges, (params: any, [start, end]: any) => {
            const comment = params[0];
            const onString = comment.slice(0, end - start);
            const notOnString = comment.slice(end - start - 1, comment.length);
            return onString + this.clearComments(notOnString);
        }, () => ""));
    }

    public parseQueryCode(code: string) {
        const doubleWhiteSpaceRegex = /\s+/g;
        const commentless = this.clearComments(code);
        const stringRanges = this.extractStringsRangeFromQuery(commentless);
        const parsedCode = commentless.replace(doubleWhiteSpaceRegex, this.strReplaceNotIn(stringRanges, (params: any) => params[0], () => " "));
        return parsedCode.trim();
    }
    //  Metodos do Form
    public fillDataSourceForm({ row }: any) {
        if (!row) {
            return;
        }
        const values = row || this.datasourceData;
        const form = Metadata.getInstance<Form>("dataSourceForm");
        const query = this.isGridVisible ? this.addDatasourceQuery : this.editDatasourceQuery;
        const accessProfiles = form.getChildInstance<SelectMultiple>("accessProfiles");
        form.value = values;
        query.code = values.query_view || values.query || "";
        accessProfiles.setValue(values.accessProfiles || []);
    }

    public async saveDatasourceForm() {
        const form = this.isGridVisible ? Metadata.getInstance<Form>("modalDataSourceForm") : Metadata.getInstance<Form>("dataSourceForm");
        const query = this.isGridVisible ? this.addDatasourceQuery : this.editDatasourceQuery;
        const values = form.value;
        const connection = values.connection;
        const queryCode = query.code ? this.parseQueryCode(query.code) : "";
        if (!form.validate() || !queryCode.trim()) {
            if (!queryCode.trim()) {
                CustomAlertService.errorAlert("O campo de código é obrigatório e deve ser preenchido");
            }
            return;
        }
        const data = {
            ...values,
            changed: true,
            columns: this.datasourceData.columns || [],
            connection: connection?._id,
            connectionName: connection?.name,
            room: connection?.room,
            createdBy: this.datasourceData.userId,
            importStatus: this.datasourceData.importStatus || "",
            originDataSources: this.datasourceData.originDataSources || "",
            query: queryCode,
            query_view: query.code || "",
            relations: this.datasourceData.relations || [],
            _queryChanged: this.datasourceData._id ? !(this.datasourceData.query === queryCode) : true,
            _id: this.datasourceData?._id
        };
        try {
            const res = await this.saveDataSource(data);
            CustomAlertService.saveSuccessAlert();
            if (res.data.messages?.length > 0) {
                const message = res.data.messages[0].message;
                DialogService.show({
                  name: 'warning-dialog',
                  type: 'error',
                  title: ' ',
                  text: message
                });
            }
            this.reloadGrid();
            this.closeDetails();
        } catch (err) {
            LoadingService.hide();
            CustomAlertService.errorAlert("Ocorreu um erro.");
            console.error(err);
        }
    }
    public async saveDataSource(datasourceData: any) {
        LoadingService.show();
        const res = await Http.post("/dataSource/save", {
            requestType: "DataSet",
            disableLoader: false,
            dataset: [datasourceData._id ? datasourceData : {
                ...datasourceData,
                __createdLocal: true,
                __is_new: true,
                createdBy: this.user.userId
            }]
        });
        LoadingService.hide();
        return res;
    }

    protected user: any;

    public async updateLoggedUserInfo() {
        const appController = Loader.getInstance('AppController');
        const user = await appController.loginPromise.then();
        const login = user;
        this.user = {
            userId: login.userId,
            username: login.username
        };

        const grid = Metadata.getInstance<TreeGrid>('datasource-grid');
        await grid.reload();

    }

    public handleConnectionResponse({ component }: any) {
        const info = component?.currentRow;
        if (info?.WALLET_NRORG === undefined) {
            return;
        }
        const {data, WALLET_NRORG, pagination} = info;
        component.data = data;
        component.total = pagination.total;
        component.limit = pagination.limit;
        component.page = pagination.page;
        component.currentRow = data;
    }
    public reloadGrid() {
        const grid = Metadata.getInstance<TekGrid>("datasource-grid");
        grid.reload();
    }
    public async saveInlineEdit(params: any) {
        Utilities.genericBeforeSaveTekGrid(params, "dataSource/save", "datasource-grid");
    }

    public adjustFilter({ component }: any) {
        const datasource = component.datasource;
        const filter = datasource.dynamicFilter;
        Object.entries(filter).forEach(([k, filters]: any) => {
            filter[k] = filters.filter(({value}: any) => !(value instanceof Array && value.length === 0));
            if (filter[k].length === 0) {
                delete filter[k];
            }
        });
    }
}
